/**
 *
 * $Id: ansi.cpp,v 1.21 2007/10/26 03:26:09 mviara Exp $
 * $Author: mviara $
 * $Name:  $
 *
 * ANSI Terminal emulator derived from ANSIX364 written 15 Feb 1992.
 *
 * $Log: ansi.cpp,v $
 * Revision 1.21  2007/10/26 03:26:09  mviara
 * Improved performance of function keys.
 *
 * Revision 1.20  2006/01/09 11:51:41  mviara
 * Version 2.07
 *
 * Revision 1.19  2006/01/05 06:24:05  mviara
 * Fixed bug in ESC ( ...
 *
 * Revision 1.18  2005/11/21 19:00:02  mviara
 * Added supporto to report terminal id and windows title.
 *
 * Revision 1.17  2005/05/31 17:37:26  mviara
 * Fixed bug in telnet option NAWS
 *
 * Revision 1.16  2005/05/18 11:07:55  mviara
 * Added support for  RFC1093 (Windows size)
 *
 * Revision 1.15  2005/04/19 14:05:37  mviara
 * Added supporto save restore attribute with ESC7 / ESC8
 *
 * Revision 1.14  2005/04/03 08:27:47  mviara
 * Added support for copy paste text.
 *
 * Revision 1.13  2005/04/02 07:26:29  mviara
 * Added partial support for copy/paste text.
 *
 * Revision 1.12  2005/03/28 12:53:09  mviara
 * Added support for xterm, fixed bug in attributr at begin of line.
 *
 * Revision 1.11  2005/01/24 17:31:21  mviara
 * Fixed bug in handling ESC[m
 *
 * Revision 1.10  2004/12/06 22:12:13  mviara
 * Removed unused code.
 *
 * Revision 1.9  2004/12/06 20:18:16  mviara
 * Added some English comment.
 *
 * Revision 1.6  2004/11/30 20:06:41  mviara
 * Added support for hide/show cursor.
 * Set default to green on black.
 *	
 * Revision 1.5  2004/11/10 18:40:17  mviara
 * Version 2.01
 *	
 *	
 * Revision 1.3  2004/07/10 14:58:18  mviara
 * Correct error in tab at end of screen.
 *	
 * Revision 1.2  2004/07/10 13:59:58  mviara
 * Add support for connection time.
 *	
 * Revision 1.1.1.1  2004/07/10 12:33:21  mviara
 * Migrated from RCS
 *	
*/	
#include "ansi.hpp"

static char rcsinfo[]= "$Id: ansi.cpp,v 1.21 2007/10/26 03:26:09 mviara Exp $";



/**
 * Scan code of funcion key
 */
unsigned char scancode[] =
{
	0x3b,		// f1    0
	0x3c,		// f2	 1
	0x3d,		// f3	 2
	0x3e,		// f4	 3
	0x3f,		// f5	 4
	0x40,		// f6	 5
	0x41,		// f7	 6
	0x42,		// f8	 7
	0x43,		// f9	 8
	0x44,		// f10	 9
	0x85,		// f11	 10
	0x86,		// f12	 11
	0x54,		// shift f1 12
	0x55,		// shift f2 13
	0x56,		// shift f3 14
	0x57,		// shift f4 15
	0x58,		// shift f5 16
	0x59,		// shift f6 17
	0x5a,		// shift f7 18
	0x5b,		// shift f8 19
	0x5c,		// shift f9 20
	0x5d,		// shift f10 21
	0x87,		// shift f11 22
	0x88,		// shift f12 23
	0x5e,		// ctrl f1   24
	0x5f,		// ctrl f2   25
	0x60,		// ctrl f3   26
	0x61,		// ctrl f4   27
	0x62,		// ctrl f    28
	0x63,		// ctrl f6   29
	0x64,		// ctrl f7   30
	0x65,		// ctrl f8   31
	0x66,		// ctrl f9   32
	0x67,		// ctrl f10  32
	0x89,		// ctrl f11  34
	0x8a,		// ctrl f12  35
	0x5e,		// ctrl f1   36
	0x5f,		// ctrl f2   37
	0x60,		// ctrl f3   38
	0x61,		// ctrl f4   39
	0x62,		// ctrl f5   40
	0x63,		// ctrl f6   41
	0x64,		// ctrl f7   42
	0x65,		// ctrl f8   43
	0x66,		// ctrl f9   44
	0x67,		// ctrl f10  45
	0x89,		// ctrl f11  46
	0x8a,		// ctrl f12  47
	0x47,		// home	     48
	0x48,		// up arrow  49
	0x49,		// page up   50
	0x4a,		// -	     51
	0x4b,		// left arrow 52
	0x4c,		// 5		53
	0x4d,		// right arrow	54
	0x4e,		// +		55
	0x4f,		// end		56
	0x50,		// down arrow	57
	0x51,		// page down	58
	0x52,		// ins		59
	0x53,		// del		60
	0x68,		// alt F1	61
	0x69,		// alt F2	62
	0x6a,		// alt F3	63
	0x6b,		// alt F4	64
	0x6c,		// alt F5	65
	0x6d,		// alt F6	66
	0x6e,		// alt F7	67
	0x6f,		// alt F8	68
	0x70,		// alt F9	69
	0x71,		// alt F10	70

};

int ScanCodeIndex(unsigned char k)
{
	unsigned char *p = (unsigned char *)memchr(scancode,k,sizeof(scancode));

	return p == 0 ? -1 : p - scancode;
}

void CAnsi::PrintScreen()
{
	SCHAR * ptr;

	ptr = curScr->buffer;
	AnsiPrintf("\33[5i");

	for (int r = 0 ; r < numRows ; r++)
	{
		for (int c = 0 ; c < numCols ; c++,ptr++)
			DisplayPrinter(ptr->chr);
		AnsiPrintf("%c%c",13,10);
	}
	AnsiPrintf("%c",12);
	AnsiPrintf("\33[4i");
}

void CAnsi::Info()
{
	int i;
	struct tm * tp;
	time_t curTime;
	double connectTime;
	
	int oldScreen = SwitchScreen(MAXSCREENS);
	HideCursor();
	
	AnsiPrintf("\33[4;1H");
	AnsiPrintf("Adapter : %s (%dx%d)\n",adapter->Name(),
		   numCols,numRows);
	AnsiPrintf("Device  : %s (Read %ld, Written %ld)\n",
		   device->Name(),deviceRead,deviceWrite);
	AnsiPrintf("Printer : %s\n",printerName);
	AnsiPrintf("Switch  : %d (%d)\n",flagSwitch,MAXSCREENS);
	AnsiPrintf("Capture : %s\33[K\n",fdCapture == -1 ? "off" : "on ansi.cap");
	AnsiPrintf("Copy    : %5d char\n",strlen(copyText));
	AnsiPrintf("\n");
	time(&curTime);
	
	for (i = 0 ; i < MAXSCREENS; i++)
	{
		AnsiPrintf("Session # %2d ",i+1);
		
		if (screens[i] != 0)
		{
			tp = localtime(&screenTime[i]);
			connectTime = difftime(curTime,screenTime[i]);
			AnsiPrintf("Connected     %02d/%02d/%04d %02d:%02d:%02d",
				   tp->tm_mday,tp->tm_mon+1,tp->tm_year+1900,
				   tp->tm_hour,tp->tm_min,tp->tm_sec);
			AnsiPrintf(" Elapsed %10.0f sec",connectTime);
				   
		}
		else
			AnsiPrintf("Not connected");
		
		AnsiPrintf("\n");


	}
	
	ShowCursor();
	
	AnsiPrintf("Press any key to continue");
	KeybRead();
	AnsiPrintf("\n\n");
	
	SwitchScreen(oldScreen);
	
}


int CAnsi::Read(char * buffer,int n)
{
	int i ;

	if (fdFromFile >= 0)
	{
		i = read(fdFromFile,buffer,n);
		if (i <= 0)
		{
			close(fdFromFile);
			fdFromFile = -1;
		}
		else
			return i;
	}
	
	if ((i = device->Read(buffer,n)) < 0)
	{
		flagExit = 1;
		i = 0;
	}

	if (i && !endConnect[curScr->session])
	{
		endConnect[curScr->session] = ::GetTickCount();
		AnsiPrintf(" (%lu ms)\n\n",endConnect[curScr->session] - startConnect[curScr->session]);
	}
	deviceRead += i;
	
	return i;
}

int CAnsi::Write(char * buffer,int n)
{
	int i;

	if ((i = device->Write(buffer,n)) != n)
	{
		flagExit = 1;
		i = 0;
	}
	
	deviceWrite += i;

	return i;
}

void CAnsi::KeybInsert(unsigned k)
{
	//
	//	Se il buffer e' pieno torna errore
	//
	if (((curScr->ps+1) & CScreen::KEYBUF) == curScr->pl)
		return;

	curScr->keybuf[curScr->ps++] = (unsigned char)k;

	curScr->ps &= CScreen::KEYBUF;

	return;
}

void CAnsi::KeybInsert(char * p)
{
	while (*p)
		KeybInsert(*p++);
}

void CAnsi::DeleteScreen()
{
	if (curScr)
	{
		device->Disconnect(curScr->session);
		screens[curScr->session] = 0;
		delete curScr;
		curScr = 0;
	}
}

int CAnsi::SwitchScreen(int s)
{
	int oldScreen;


	if (curScr == 0)
		oldScreen = -1;
	else
		oldScreen = curScr->session;
	
	if (curScr == 0 || s != curScr->session)
	{
		if (curScr != 0)
			DeactivateScreen(curScr->session);
		
		
		if (screens[s] == 0)
		{
			AddScreen(s);
		}
		else
		{
				
			ActivateScreen(s);
			if (s < MAXSCREENS)
				ConnectScreen(s);
		}
	}

	
	return oldScreen;
}

void CAnsi::PasteText()
{
	for (int i = 0 ; copyText[i] != 0 ; i++)
		KeybInsert(copyText[i]);
}

void CAnsi::CopyText()
{
	int pos = curScr->curPos;
	int exit = 0;
	SCHAR * copyBuffer = new SCHAR[screenSize];
	int startPos = -1,endPos;
	char * oldStatusLine = new char[numCols];
	
	memcpy(copyBuffer,curScr->buffer,screenSize*sizeof(SCHAR));
	memcpy(oldStatusLine,curScr->statusLine,numCols);
	
	adapter->BigCursor();
	StatusLine("Move to start and press ENTER");
	
	while (!exit)
	{
		Locate();
		int key = adapter->Getch();
		switch (key)
		{
			default:
					if (startPos == -1)
					{
						StatusLine("Move to end and press ENTER");
						startPos = curScr->curPos;
						endPos   = startPos;
					}
					else
						exit = 1;
					break;
					
			case	27:
					exit = 1;
					break;
					
			case	0x4b00:
					if (curScr->curPos)
						curScr->curPos--;
					break;

			case	0x4700:
					curScr->curPos = 0;
					break;
			case	0x4f00:
					curScr->curPos = screenSize -1;
					break;

			case	0x4d00:
					if (curScr->curPos + 1 < screenSize)
						curScr->curPos++;
					break;
					
			case	0x4800:
					if (curScr->curPos >= numCols)
						curScr->curPos -= numCols;
					break;
					
			case	0x5000:
					if (curScr->curPos + numCols < screenSize)
						curScr->curPos += numCols;
		}

		if (startPos != -1)
		{
			endPos = curScr->curPos;
			int r1,c1,r2,c2,r,c;
			r1 = startPos / numCols;
			c1 = startPos % numCols;
			r2 = endPos / numCols;
			c2 = endPos % numCols;
			if (r2 < r1)
			{
				r = r2;
				r2 = r1;
				r1 = r;
			}

			if (c2 < c1)
			{
				c = c2;
				c2 = c1;
				c1 = c;
			}

		
			int count = 0;
			for (r = 0 ; r < numRows ; r++)
			{
				int p = r * numCols;
				for (c = 0 ; c < numCols ; c++,p++)
				{
					SCHAR ch = copyBuffer[p];

					if (r >= r1 && r <= r2 && c >= c1 && c <= c2)
					{
						copyText[count++] = ch.chr;
						if (c == c2 && r < r2)
							copyText[count++] = 13;
						
						int f = ch.color & 7;
						int b = (ch.color & 0x70) >> 4;
						ch.color = (ch.color & 0x88) | (f << 4) | b;
					}
					VideoPut(p,1,&ch);
					
				}
				copyText[count] = 0;
			}

			adapter->BigCursor();

		}
	}

	
	adapter->LittleCursor();
	VideoPut(0,screenSize,copyBuffer);
	memcpy(curScr->statusLine,oldStatusLine,numCols);
	ShowStatusLine();
	curScr->curPos = pos;
	Locate();
	delete oldStatusLine;
	delete [] copyBuffer;
}

void CAnsi::KeybInput()
{
	unsigned k;
	register int i;
	int	j;

	k=adapter->Getch();

	//AnsiPrintf("Getch = %x\n",k);
	if (k < 0x100)
	{
		KeybInsert(k);
		return;
	}

	switch (k)
	{
		case	KEYSCROLLOFF:
			curScr->scrolloff = 1;
			break;

		case	KEYSCROLLON:
			curScr->scrolloff = 0;
			break;
#define ALT(x) KeybInsert(27);KeybInsert(x);return
		case	0x1e00:	// ALT-A
				ALT('A');
				
		case	0x3000:	// ALT-B
				ALT('B');

		case	0x2E00:	// ALT-C
				ALT('C');
				
		case	0x2000:	// ALT-D
				ALT('D');

		case	0x1200:	// ALT-E
				ALT('E');

		case	0x2100:	// ALT-F
				ALT('F');

		case	0x2200:	// ALT-G
				ALT('G');

		case	0x2300:	// ALT-H
				ALT('H');

		case	0x1700:	// ALT-I
				ALT('I');

		case	0x2600:	// ALT-L
				ALT('L');

		case	0x3200:	// ALT-M
				ALT('M');

		case	0x3100:	// ALT-N
				ALT('N');

		case	0x1800:	// ALT-O
				ALT('O');

		case	0x1900:	// ALT-P
				ALT('P');

		case	0x1000:	// ALT-Q
				ALT('Q');

		case	0x1300:	// ALT-R
				ALT('R');

		case	0x1F00:	// ALT-S
				ALT('S');

		case	0x1400:	// ALT-T
				ALT('T');

		case	0x2D00:	// ALT-X
				ALT('X');

		case	0x1500:	// ALT-Y
				ALT('Y');

		case	0x2F00:	// ALT-V
				ALT('V');

		case	0x1100:	// ALT-W
				ALT('W');

		case	0x2C00:	// ALT-Z
				ALT('Z');

		//case	
//		case	0x2300:	/*	ALT-H	*/
//			help();
//			return;
				
		case	0x1701:	/*	CTRL-ALT-I	*/
			Info();
			return;

		case	0x1901:	// CTRL-ALT-P Print Screen
			PrintScreen();
			return;
			
		case	0x2e01: // CTRL-ALT-C Capture
			Capture();
			return;

			
		case	0x3201:	// CTRL-ALT-M
			Macro();
			return;
		
		case	0x6800:
		case	0x6900:
		case	0x6a00:
		case	0x6b00:
		case	0x6c00:
		case	0x6d00:
		case	0x6e00:
		case	0x6f00:
		case	0x7000:
		case	0x7100:
				k >>= 8;
				k -= 0x68;
				if (k >= MAXSCREENS)
					break;
				if (flagSwitch)
					SwitchScreen(k);
				return;
								
		case	0x2d01: /*	CTRL-ALT-X	*/
				flagExit = 1;
				return;

		case	0x5200:
				if (adapter->GetState().flag.ctrl)
				{
					CopyText();
					return;
				} else if (adapter->GetState().flag.shift)
				{
					PasteText();
					return;
				}
				break;
			
				
	}

	/* codice dello scan code nella parte bassa */
	k >>= 8;

/*
	for (i = 0 ; i < CScreen::FCNKEY ; i++)
		if (k == scancode[i])
		{
			for (j = 0 ; j < curScr->lenght[i];j++)
			{
				KeybInsert(curScr->keys[i][j]);
			}
			return;
		}
*/

	i = ScanCodeIndex(k);
	if (i != -1)
	{
		for (j = 0 ; j < curScr->lenght[i];j++)
		{
			KeybInsert(curScr->keys[i][j]);
		}
		return;
		
	}
	
	/*
	 * Se non trova il tasto funzionale lo inserisce normale
	*/
	KeybInsert(0xe0);
	KeybInsert(k);
}

int CAnsi::KeybStatus()
{
	
	KSTATE newState;
	/* loop finche ci sono caratteri */
	while (adapter->Kbhit())
		KeybInput();

	newState = adapter->GetState();
	if (newState.flags != curScr->state.flags)
	{
		curScr->state.flags = newState.flags;
		ShowStatusLine();
	}
	
	if (curScr->pl == curScr->ps)
		return 0;
	else
		return 1;
}


int CAnsi::KeybRead()
{
	int c;

	while (!KeybStatus())
		KeybInput();

	c = curScr->keybuf[curScr->pl++];
	curScr->pl &= CScreen::KEYBUF;

	return c & 0xff;
}


int CAnsi::GetPar(char c)
{
	if (c >= '0' && c <= '9')
	{
		if (curScr->nopar)
		{
			*curScr->parptr =
			curScr->nopar = 0;
		}
		*curScr->parptr *= 10;
		*curScr->parptr += c - '0';
		return 0;
	}

	if (c == ';')
	{	
		if (++curScr->numpar >= (int)(sizeof(curScr->par)/sizeof(int)))
		{
			curScr->numpar--;
			curScr->parptr --;
		}
		curScr->parptr ++;
		curScr->nopar = 1;
		return 0;
			
	}

	return 1;
	
}

void CAnsi::Delchar(int col,int no,int eol)
{
	int source;
	int dest;
	unsigned count;
	unsigned current;
	int ssize;
	int row = curScr->curPos / numCols;

	
	if (eol)
		ssize = numCols*(row+1);
	else
		ssize = numCols * numRows;

	current = row * numCols + col;
	dest = row*numCols+col;
	source = dest + no;
	count = ssize - current - no;
	VideoCopy(source,dest,count);
	VideoSet(ssize-no,' ',no);
}

void CAnsi::Inschar(int col,int no,int eol)
{
	int  source;
	int  dest;
	unsigned count;
	unsigned current;
	int ssize;
	int row = curScr->curPos / numCols;
	
	if (eol)
		ssize = numCols*(row+1);
	else
		ssize = numCols * numRows;

	current = row * numCols + col;

	source = row*numCols+col;
	dest = source + no;
	count = ssize - current	- no;
	VideoCopy(source,dest,count);
	VideoSet(source,' ',no);
}

unsigned char CAnsi::tabColor[] =
{
	0,4,2,6,1,5,3,7	// Black,Red,Green,Yellow,Blue,Magenta,Cyan,White
};

unsigned char CAnsi::Vga2Ansi(unsigned char c)
{
	c &= 0x07;
	for (int i = 0 ; i < sizeof(tabColor) ; i++)
		if (c == tabColor[i])
			return i;
	return 0;
}

unsigned char CAnsi::Ansi2Vga(unsigned char c)
{
	return tabColor[c & 0x07];
}


void CAnsi::ProcessAttribute(int n,int num)
{
	
	switch (n)
	{
		case	0:
				curScr->attribute.bcolor = curScr->attribute.bnormal;
				curScr->attribute.fcolor = curScr->attribute.fnormal; 
				curScr->attribute.hi =
				curScr->attribute.blink = 
				curScr->attribute.reverse = 0;
				break;

		case	1:
				curScr->attribute.hi = 1;
				break;
				
		case	4:	/* underscore */

				break;

		case	5:
				curScr->attribute.blink = 1;
				break;

		case	7:
				curScr->attribute.reverse = 1;
				break;

		case	8:
				curScr->attribute.fcolor = curScr->attribute.bcolor = 0;
				break;

				/* font 1 */
		case	10:
				curScr->orMask = 0;
				break;

		case	12:	/* font 2 */
				curScr->orMask = 0x80;
				break;

		case	27:
				curScr->attribute.reverse = 0;
				break;
				
				/* foreground color */
		case	30:	// Black
		case	31:	// Red
		case	32:	// Green
		case	33:	// Yellow
		case	34:	// Blue
		case	35:	// Magenta
		case	36:	// Cyan
		case	37:	// White
				curScr->attribute.fcolor =
				curScr->attribute.freverse = Ansi2Vga(n - 30);
				break;
			
		case	38:	/* underline on */
		case	39:	curScr->attribute.fcolor =
				curScr->attribute.freverse = curScr->attribute.fnormal;
				break;
		case	40:
		case	41:
		case	42:
		case	43:
		case	44:
		case	45:
		case	46:
		case	47:
				curScr->attribute.bcolor =
				curScr->attribute.breverse = Ansi2Vga(n - 40);
				break;
		case	49:
				curScr->attribute.bcolor =
				curScr->attribute.breverse = curScr->attribute.bnormal;
			break;
	}

	CalcAttribute();
}

void CAnsi::CsiAttribute()
{
	int i;

	if (curScr->numpar == 0 && curScr->nopar)
	{
		ProcessAttribute(0,0);
		return;
	}
	if (curScr->nopar)
		curScr->numpar--;
	//for (i = 0 ; curScr->parptr >= &curScr->par[i] ; i ++)
	for (i = 0 ; i <= curScr->numpar ; i ++)
	{
		ProcessAttribute(curScr->par[i],curScr->numpar);
	}
}

void CAnsi::SetColor(unsigned c)
{
	curScr->newColor = c;
	curScr->curColor = ~c;
	SetColor();
}

void CAnsi::SetColor()
{
	if (curScr->newColor != curScr->curColor)
	{
		curScr->curColor = curScr->newColor;
		curScr->SetColor(curScr->curColor);
		adapter->SetColor(curScr->curColor);
	}
};

void CAnsi::CalcAttribute()
{
	if (curScr->attribute.reverse)
		curScr->newColor = (curScr->attribute.freverse << 4) | curScr->attribute.breverse;
	else
		curScr->newColor = (curScr->attribute.bcolor << 4) | curScr->attribute.fcolor;

	if (curScr->attribute.blink)
		curScr->newColor |= 0x80;
	if (curScr->attribute.hi)
		curScr->newColor |= 0x08;

	//SetColor(curScr->curColor);
}

void  CAnsi::PrinterOutput(unsigned char c)
{
	
	if (curScr->fdlpt == NULL)
	{
		char *p = printerName;
		
		if (doubleBuffer)
		{
			p = tmpnam(NULL);
			strcpy(curScr->lptTmp,p);
		}
		
		curScr->fdlpt = fopen(p,"ab");
		
		if (curScr->fdlpt == (FILE *)NULL)
		{
			printf("ANSI32: Unable to open %s\n",
				   p);
			adapter->Getch();
			strcpy(printerName,"nul");
			return ;
		}
	}

	putc(c,curScr->fdlpt);
}


void CAnsi::StopPrinter()
{

	
	VideoPut(0,screenSize,curScr->lptVideoBuffer);
	delete [] curScr->lptVideoBuffer;
	curScr->lptVideoBuffer = 0;
	curScr->curPos = curScr->lptCursor;
	ShowCursor();
	
	fflush(curScr->fdlpt);
	fclose(curScr->fdlpt);
	
	if (doubleBuffer)
	{
		char buffer[128];
		sprintf(buffer,"copy %s %s/b>nul",curScr->lptTmp,
			printerName);
		system(buffer);
		unlink(curScr->lptTmp);
	}
	
	curScr->fdlpt = NULL;
	curScr->DisplayChar = DisplayNormal;
	
}

void CAnsi::SelectPrinter(char *name)
{
	int c = 0;
	unsigned k;
	

	if (!selectPrinter)
		return;

	for (;;)
	{
		AnsiPrintf("\033[09;39H\033[1m%-32.32s\033[0m",name);
		AnsiPrintf("\033[09;%dH",39+c);
		k = adapter->Getch();

		if (k == 13 || k == 10 || k == 27)
			break;
		
		if (k > 32 && k <= 128)
		{
			if (c < 32)
			{
				if (name[c] == 0)
					name[c+1] = 0;
				name[c] = (char)k;
				c++;
			}
		}

		if (k == 0x5300)
		{
			if (name[c] != 0)
				strcpy(name+c,name+c+1);
		}

		if (k == 0x4d00)
		{
			if (name[c] != 0)
				c++;
		}

		if (k == 0x4b00)
		{
			if (c)
				c--;
		}

		if (k == 8)
		{
			if (c)
			{
				strcpy(name+c-1,name+c);
				c--;
			}
		}
		
		
	}

	if (!strlen(name))
		strcpy(name,"nul");
	
		
}

void CAnsi::StartPrinter()
{
	curScr->lptVideoBuffer = new SCHAR[screenSize];
	curScr->lptCursor = curScr->curPos;
	assert(curScr->lptVideoBuffer != 0);
	memcpy(curScr->lptVideoBuffer,curScr->buffer,sizeof(SCHAR)*screenSize);

	HideCursor();
	ResetScreen();
	
	AnsiPrintf("\033[09;27HDevice :");
	SelectPrinter(printerName);
	AnsiPrintf("\033[09;39H%s",printerName);
	AnsiPrintf("\033[10;25H* Please wait .. printing *");
	
	curScr->DisplayChar = DisplayPrinter;
	curScr->lptCounter = 0;
	curScr->printchar = 0;
	curScr->fdlpt = NULL;
	adapter->Kbhit();
}

void CAnsi::DisplayPrinter(unsigned char c)
{
	register int i;
	
	static char end_print[] = "\33[4i";/* sequenza di fine stampa	*/
	
	//
	// Se fa parte della fine stampa vede se va in normale
	//
	if (c == end_print[curScr->printchar])
	{
		/* se e' un fine stampa termina */
		if (++curScr->printchar == sizeof(end_print) -1 )
		{
			StopPrinter();
		}
	}
	else
	{
		if ((++curScr->lptCounter % 1024) == 0)
		{
			curScr->DisplayChar = DisplayNormal;
			
			AnsiPrintf("\033[11;27HPrinted %8lu byte(s)\n",
				   curScr->lptCounter);
			adapter->Kbhit();
			curScr->DisplayChar = DisplayPrinter;
			
		}
		
		if (curScr->printchar)
		{
			for (i = 0 ; i < curScr->printchar ; i++)
				PrinterOutput(end_print[i]);
			
			curScr->printchar = 0;
			
			/* richiama se stesso perche potrebbe comiciare */
			/* una nuova seqenza di escape !!!!		*/
			DisplayPrinter(c);
		}
		else
			PrinterOutput(c);
			
	}
}

void  CAnsi::DisplayVt100Mode(unsigned char c)
{
	switch (c)
	{
		case	'0':
			curScr->gvt100 = 1;
			break;
		case	'A':
		case	'B':
			curScr->gvt100 = 0;
			break;
			break;
#ifndef NDEBUG
		default:
			FILE * fdx = fopen("unknown.txt","a");
			fprintf(fdx,"ESC ( %c\n",c);
			fclose(fdx);
			break;
#endif
	}
	
	curScr->DisplayChar = DisplayNormal;
		
}

void  CAnsi::DisplayAdditional(unsigned char c)
{
	char buffer[33];

	/*
	 *	Se sono parametri ha finito
	*/
	if (!GetPar(c))
		return;

	switch (c)
	{
		default:
			OutChar(27);
			OutChar('[');
			OutChar('=');
			OutChar(c);
			break;
			
			/* output fcolor bcolor */
		case	'M':
			KeybInsert(itoa(Vga2Ansi(curScr->attribute.fcolor),buffer,10));
			KeybInsert(' ');
			KeybInsert(itoa(Vga2Ansi(curScr->attribute.bcolor),buffer,10));
			KeybInsert(13);
			break;
			
			/* set cursor start end line */
		case	'C':
			if (curScr->numpar == 0 || curScr->par[0] > 10)
			{
				//curScr->cursorOn = 1;
				HideCursor();
			}
			else
			{

				//curScr->cursorOn = 0;
				ShowCursor();
			}
			break;

			/* attiva disattiva hi intensiti */
		case	'D':
			if (curScr->par[0])
				curScr->attribute.hi = 1;
			else
				curScr->attribute.hi = 0;
			break;

			/* attiva disattiva blink */
		case	'E':
			if (curScr->par[0])
				curScr->attribute.blink = 1;
			else
				curScr->attribute.blink = 0;
			break;

			/* foreground normal color */
		case	'F':
			curScr->attribute.fnormal	= Ansi2Vga(curScr->par[0]);
			break;

			/* background normal color */
		case	'G':
			curScr->attribute.bnormal = Ansi2Vga(curScr->par[0]);
			break;

			/* reverse foreground */
		case	'H':
			curScr->attribute.freverse = Ansi2Vga(curScr->par[0]);
			break;

			/* reverse background */
		case	'I':
			curScr->attribute.breverse = Ansi2Vga(curScr->par[0]);
			break;

		case	'J':
			curScr->attribute.fgraphics = curScr->par[0]; // 15.01.98
			break;

		case	'K':
			curScr->attribute.bgraphics = curScr->par[0]; // 15.01.98
			break;

	}

	CalcAttribute();
	curScr->DisplayChar = DisplayNormal;
}

void CAnsi::DisplaySpecial(unsigned char c)
{
	int i;
	int mode;

	if (!GetPar(c))
		return;
	
	curScr->DisplayChar = DisplayNormal;

	if	(c == 'h')
		mode = 1;
	else if (c == 'l')
		mode = 0;
	else
	{
		OutChar(27);
		OutChar('[');
		OutChar('?');
		OutChar(c);
		return;
				
	}
	
	for (i = 0 ; curScr->parptr >= &curScr->par[i] ; i ++)
		switch (curScr->par[i])
		{
			case	3:	// 132 /80 colonne
				break;
			case	7:
				curScr->autoWrap = !mode;
				break;
				
			case	25:	/* cursor on / off */
				if (mode)
					ShowCursor();
				else
					HideCursor();
				break;
		}
}

void CAnsi::DisplayCsi(unsigned char c)
{
	char tmp[320];
	
#ifndef NDEBUG
	FILE *fdx;
	int i;
#else
	int i;
#endif
	
	int oldcol,oldrow;
	int row,col;


	if (curScr->csi == ']')
	{
		if (curScr->numpar == 0)
			if (!GetPar(c))
				return;
		
		if (c != 7)
		{
			curScr->pt[curScr->ptlen++] = c;
			
		}
		else
		{
			curScr->pt[curScr->ptlen] = 0;
			StatusLine("%s",curScr->pt);
			curScr->DisplayChar = DisplayNormal;
		}
		return;
	}

	if (!GetPar(c))
		return;
	
	col = oldcol = curScr->curPos % numCols;
	row = oldrow = curScr->curPos / numCols;

	switch (c)
	{
		default:
#ifndef NDEBUG
			fdx = fopen("unknown.txt","a");
			fprintf(fdx,"ESC [ ");
			for (i = 0 ;i < curScr->numpar ; i++)
			{
				fprintf(fdx,"%d",curScr->par[0]);
				if (i+1 < curScr->numpar)
					fprintf(fdx,";");
			}
			fprintf(fdx,"%c\n",c);
			fclose(fdx);
#endif
			OutChar(27);
			OutChar('[');
			OutChar(c);
			break;

		case	'G':
		case	'`':
				col = curScr->par[0] - 1;
				break;
			
		case	'?':
				StartCsi();
				curScr->DisplayChar = DisplaySpecial;
				return;
				
		case	'r': // vt100
				curScr->bscroll = curScr->par[0] - 1;
				curScr->escroll = curScr->par[1] - 1;
				break;
			
				// back tab
		case	'Z':
				BackTab(curScr->par[0]);
				break;
			

			
				/* CSI= ... sequenze ansi addizionali */
		case	'=':
				curScr->DisplayChar = DisplayAdditional;
				return;

				/* insert line */
		case	'L':
				if (curScr->nopar)
					curScr->par[0] = 1;
				Inschar(0,curScr->par[0]*numCols,0);
				break;

				/* delete line */
		case	'M':
				if (curScr->nopar)
					curScr->par[0] = 1;
				Delchar(0,curScr->par[0]*numCols,0);
				break;

				/* delete char */
		case	'P':
				if (curScr->nopar)
					curScr->par[0] = 1;

				Delchar(col,curScr->par[0],1);
				break;
				
				/* insert char */
		case	'@':
				if (curScr->nopar)
					curScr->par[0] = 1;

				Inschar(col,curScr->par[0],1);
				break;
			
		case	'i': /* gestione della stampante */
			if (curScr->par[0] == 5)
			{
				StartPrinter();
				return;
			}
			break;
			
		case	'm':
			/* gestione attributi */
			CsiAttribute();
			break;
		case	'n':
			switch (curScr->par[0])
			{
				case	6:
					sprintf(tmp,"\33[?%d;%dR",row+1,col+1);
					KeybInsert(tmp);
					break;
			}
			break;
			
			/* spostamenti del cursore */
		case	'A':
			curScr->wrapped = 0;
			row -= curScr->par[0];
			break;

		case	'B':
		case	'e':
			row += curScr->par[0];
			break;

		
		case	'C':
		case	'a':
			col += curScr->par[0];
			break;
		case	'D':
			col -= curScr->par[0];
			break;

			/* cambia colonna del cursore */
		case	'\'':
			col = curScr->par[0] - 1;
			break;

			
		case	'f':
		case	'H':
			row = curScr->par[0] - 1;
			col = curScr->par[1] - 1;
			curScr->wrapped = 0;
			break;

			/* cambio riga corrente */
		case	'd':
				row = curScr->par[0] - 1;
				break;

				/* cursore inizio linea in su */
		case	'F':
				col = 0;
				row -= curScr->par[0];
				break;

				/* cursore inizio linea in giu */
		case	'E':
				col = 0;
				row += curScr->par[0];
				break;
				
				/* save cursor */
	
		case	's':
				curScr->oldCurPos = curScr->curPos;
				break;
				
				/* restore cursor */
		case	'u':
				row = curScr->oldCurPos / numCols;
				col = curScr->oldCurPos % numCols;
				break;
		case	'c':
				sprintf(tmp,"\033[0;%s;0;c",versionString);
				KeybInsert(tmp);
				break;
				
		case	't':	// 22 November 2005
				for (i = 0 ; i < curScr->numpar ; i++)
				{
					switch (curScr->par[i])
					{
						case	21:
							sprintf(tmp,"\033]2;%s\007",curScr->statusLine);
							KeybInsert(tmp);
							break;
					}
				}
				break;
				
				/* scroll up */
		case	'S':
				ScrollUp(curScr->par[0]);
				break;

				/* scroll down */
		case	'T':
				ScrollDn(curScr->par[0]);
				break;

		case	'X': /* cancella i caratteri */
				while (curScr->par[0]--)
					OutChar(' ');
				col = oldcol;
				row = oldrow;
				oldcol = -1;
				break;

		case	'K':	/* cancella linea	*/
				SetColor();
				if (curScr->nopar)
					curScr->par[0] = 0;

				switch (curScr->par[0])
				{
							/* clear to EOL */
					case	0:
							if (curScr->wrapped == 0)
								VideoClear(row,col,1,numCols-col);
							break;
							
					case	1:	/* clear to POS */
							VideoClear(row,0,1,col+1);
							//if (row)
							//	VideoClear(0,0,row-1,numCols);
							break;

					case	2:	/* clear line */
							VideoClear(row,0,1,numCols);
							break;
				}
				break;

		case	'J': /* cancella fine video ecc. */
				SetColor();
				if (curScr->nopar)
					curScr->par[0] = 0;

				switch (curScr->par[0])
				{
							/* clear EOS */
					case	0:	VideoClear(row,col,1,numCols-col);
							VideoClear(row+1,0,numRows - row-1,numCols);
							break;

					case	1:	/* clear BOS */
							VideoClear(row,0,1,numCols-col);
							VideoClear(0,0,row,numCols);
							break;

					case	2:	/* clear screen */
							ResetScreen();
							row = col = 0;
							break;
				}
				break;

		case	'z':	/* switch multi screen */
				// FIXME multi screen da fare
				break;
		case	'l':	// Reset mode
		case	'h':	// Set Mode
				// FIXME
				break;
	}


	if (oldcol != col || oldrow != row)
	{
		curScr->wrapped =0;
		
		if	(row < 0)
			row = 0;
		else if (row >= numRows)
			row = numRows - 1;

		if	(col < 0)
			col = 0;
		else if (col >= numCols)
			col = numCols - 1;

	}

	
	curScr->curPos = row * numCols + col;
	Locate();

	curScr->DisplayChar = DisplayNormal;
}
	

void CAnsi::StartCsi()
{
	curScr->ptlen = 0;
	curScr->csi = '[';
	curScr->DisplayChar = DisplayCsi;
	curScr->numpar = 0;
	curScr->nopar = curScr->par[0] =  curScr->par[1] = curScr->par[2] = 1;
	curScr->parptr = curScr->par;
}


void CAnsi::VideoSet(int pos,char ch,int n)
{
	assert(n > 0);
	assert(pos >= 0);
	assert(pos +n <= screenSize+numCols);
			
	HideCursor();
	adapter->Print(pos,ch,n);
	curScr->Print(pos,ch,n);
	ShowCursor();
}

void CAnsi::VideoClear(int row,int col,int nrow,int ncol)
{
	unsigned source;

	assert(nrow >= 0);
	assert(ncol > 0);

	
	HideCursor();
	while (nrow--)
	{
		source = row++ * numCols + col;
		if (ncol)
		{
			VideoSet(source,' ',ncol);
		}
	}

	ShowCursor();
}

void CAnsi::VideoGet(int pos,int n,SCHAR * buffer)
{
	memcpy(buffer,curScr->buffer+pos,n*sizeof(SCHAR));
}

void CAnsi::VideoPut(int pos,int n,SCHAR * buffer)
{
	char localBuffer[80];
	unsigned oldCurColor;
	int lastPos=0,num;
	SCHAR ch;

	HideCursor();
	
	oldCurColor = curScr->curColor;
	num = 0;
	ch.color =  (char)~buffer->color;
	
	while (n--)
	{
		if (ch.color != buffer->color || num >= sizeof(localBuffer))
		{
			if (num)
			{
				adapter->SetColor(ch.color);
				adapter->PrintBuffer(lastPos,localBuffer,num);
			}
			num = 1;
			lastPos = pos;
			localBuffer[0] = buffer->chr;
			ch.color = buffer->color;
		}
		else
			localBuffer[num++] = buffer->chr;
		
		curScr->buffer[pos++] = *buffer++;
		
	}

	assert(num != 0);
	
	adapter->SetColor(ch.color);
	adapter->PrintBuffer(lastPos,localBuffer,num);
	
	//SetColor(oldCurColor);
	ShowCursor();
}

void CAnsi::VideoCopy(int source,int dest,int count)
{
	assert(source >= 0 && source + count <= screenSize);
	assert(dest >=0 && dest + count <= screenSize);
	HideCursor();
	VideoGet(source	,count,tmpBuffer);
	VideoPut(dest	,count,tmpBuffer);
	ShowCursor();
}

void CAnsi::Tab()
{
	int col,row;
	int i;
	
	col =  curScr->curPos % numCols;
	row =  curScr->curPos / numCols;

	/* cerca la prima colonna superiore multipla di 8 */
	i = (col & ~ (8 - 1)) + 8;
	
	if (i < numCols)
	{
		col = i;
		curScr->curPos = row * numCols + col;
		if (curScr->curPos >= screenSize)
			curScr->curPos = screenSize - 1;
		Locate();
	}
	else
	{
		col = 0;
		curScr->curPos = row * numCols + col;
		assert(curScr->curPos < screenSize);
		Locate();
		
		LineFeed();
	}
	
}

void CAnsi::BackSpace()
{
	if (curScr->curPos % numCols)
	{
		curScr->curPos--;
		Locate();
	}
}

void CAnsi::ScrollDn(int n)
{
	int source,dest,count;
	HideCursor();
	source = curScr->bscroll * numCols;
	dest   = source + n * numCols;
	count  = (curScr->escroll - curScr->bscroll - n) * numCols;
	VideoCopy(source,dest,count);
	VideoClear(curScr->bscroll,0,n,numCols);
	ShowCursor();
	
}

void CAnsi::ScrollUp(int n)
{
	HideCursor();
	
	//
	//  Se lo scroll e' disabilitato aspetta
	//
	while (curScr->scrolloff)
		KeybInput();
	
	
	VideoCopy(numCols * (n+curScr->bscroll),curScr->bscroll*numCols,
		  (curScr->escroll - curScr->bscroll  - n + 1   ) * numCols);

	VideoClear(curScr->escroll - n+1,0,n,numCols);

	ShowCursor();
}

void CAnsi::OutChar(unsigned char c)
{
	unsigned car;

	if (c > 1)
	{

		car  = c | curScr->orMask;

		// Se e' il caso fa lo scroll
		if (curScr->curPos >= screenSize)
		{
			ScrollUp(1);
			curScr->curPos = screenSize - numCols;
			Locate();
		}

		SetColor();
		adapter->Print(curScr->curPos,(char)car,1);
		curScr->Print(curScr->curPos,(char)car,1);
		
		if (curScr->curPos % numCols == 0)
			curScr->wrapped = 1;
		else
			curScr->wrapped = 0;
		
	}

}

void CAnsi::LineFeed()
{
	int row,col;

	// Ignore line feed if the screen is wrapped
	if (curScr->wrapped)
	{
		curScr->wrapped = 0;
		return;
	}
	row = curScr->curPos / numCols;
	col = curScr->curPos % numCols;
	
	//if (++row >= numRows)
	if (++row > curScr->escroll)
	{
		row = curScr->escroll ;		
		ScrollUp(1);
	}
	curScr->curPos = row * numCols + col;
	Locate();
}

void CAnsi::CarriageReturn()
{
	int row,col;
	
	row = curScr->curPos / numCols;
	col = curScr->curPos % numCols;

	col = 0;
	//curScr->wrapped = 0;
	
	curScr->curPos = row * numCols + col;
	Locate();
}

void CAnsi::DisplayEscape(unsigned char c)
{
#ifndef NDEBUG
	FILE *fdx;
#endif
	switch (c)
	{
		case	'>': // Normal keypad
		case	'=': // Application keypad
			break;
			
		case	'(':
		case	')':
			curScr->DisplayChar = DisplayVt100Mode;
			return;

		case	'M':
			ScrollDn(1);
			break;
			
		case	'c':
			ResetScreen();	// 15.01.98
			break;
			
		case	'7':
				curScr->oldCurPos = curScr->curPos;
				curScr->oldAttribute = curScr->attribute;
				break;
			
		case	'8':
				curScr->curPos = curScr->oldCurPos;
				Locate();
				curScr->attribute = curScr->oldAttribute;
				CalcAttribute();
				SetColor();
				break;
			
				
		default:
#ifndef NDEBUG
			fdx = fopen("unknown.txt","a");
			fprintf(fdx,"ESC %c\n",c);
			fclose(fdx);
#endif
			OutChar(27);
			OutChar(c);
			break;
			
		case	'Q':
			curScr->DisplayChar = DisplayFcnkey;
			curScr->par[0] = 0;
			return;

		case	']': // Set text parameter
			StartCsi();
			curScr->csi = ']';
			return;
			
		case	'[':
			StartCsi();
			return;
			
		case	'E':
			CarriageReturn();
			//
			// Cade in line down
			//
		case	'D':
			LineFeed();
			break;

	}

	curScr->DisplayChar = DisplayNormal;

}

void CAnsi::DisplayPrgkey(unsigned char c)
{
	if (c == curScr->par[1])
		curScr->DisplayChar = DisplayNormal;
	else
	{
		if (curScr->par[0] < CScreen::FCNKEY &&
		    curScr->lenght[curScr->par[0]] < CScreen::KEYLEN)
			curScr->keys[curScr->par[0]][curScr->lenght[curScr->par[0]]++] = c;
	}
}

void CAnsi::DisplayFcnkey(unsigned char c)
{
	if (c >= '0' && c <= '9')
		curScr->par[0] = curScr->par[0] * 10 + c - '0';
	else
	{
		if (curScr->par[0] < CScreen::FCNKEY)
			curScr->lenght[curScr->par[0]] = 0;
		curScr->par[1] = c;
		curScr->DisplayChar = DisplayPrgkey;
	}
}

void CAnsi::ShowStatusLine()
{
	if (numRows < adapter->Rows())
	{
		int col = adapter->Cols();
		int c = adapter->GetColor();
		char tmp[256];
		
		memset(tmp,0304,256);
		memcpy(tmp,curScr->statusLine,strlen(curScr->statusLine));
		if (curScr->state.flag.caps)
			memcpy(tmp+col-6,"CAPS",4);
		if (curScr->state.flag.nums)
			memcpy(tmp+col-10,"NUM",3);
/*		if (curScr->state.flag.ctrl)
			memcpy(tmp+col-15,"CTRL",4);
		if (curScr->state.flag.alt)
			memcpy(tmp+col-19,"ALT",3);*/

		adapter->SetColor(11);
		adapter->PrintBuffer((adapter->Rows() - 1) * col,tmp,col - 1);
		adapter->SetColor(c);
		Locate();
	}
	
}

void CAnsi::StatusLine(char *format,...)
{
	va_list arglist;
	char buffer[256];

	va_start( arglist, format );
	vsprintf( buffer,format, arglist );
	va_end( arglist );
	assert(strlen(buffer) < sizeof(buffer));

	if (numRows < adapter->Rows())
	{
		//sprintf(curScr->statusLine,"- %s -",buffer);
		strcpy(curScr->statusLine,buffer);
		strcat(curScr->statusLine," ");
		ShowStatusLine();
	}
	    

}

void CAnsi::DisplayNormal(unsigned char c)
{
	static unsigned char vt100Tab[] =
	{	32,  4,177, 26, 23,  27, 25,248,241, 21,
		18,217,191,218,192, 197,196,196,196,196,
		196,195,180,193,194, 179,243,242,227,157,
		156,250
	};

	if (curScr->gvt100 && c >= 95 && c <= 126)
	{
		c = vt100Tab[c - 95];
	}
	
	switch (c)
	{
		case	5:
				KeybInsert(TermId());
				break;
		case	7:
				Bell();
				break;
			
		case	8:
				BackSpace();
				break;
		case	9:
				Tab();
				break;
			
		case	10:
				LineFeed();
				break;

				// Some BBS use ^L to clear the screen
		case	12: 
				ResetScreen();
				break;
				
		case	13:
				CarriageReturn();
				break;
				
		case	14:
				curScr->gvt100 = 1;
				break;
				
		case	15:
				curScr->gvt100 = 0;
				break;
		case	27:	
				StartEscape();
				break;
		case	0x9b:
				StartCsi();
				break;

		default:
				OutChar(c);
				break;
	}
}
	
void CAnsi::Putchar(char c)
{
	Outstr(&c,1);
}

void CAnsi::Outstr(char * p,int len)
{
	HideCursor();
	while (len--)
	{
		//delay(10);
		(this->*curScr->DisplayChar)(*p++);
	}
	
	ShowCursor();
}

void CAnsi::AddScreen(int s)
{
	int	i,j;
	static  char *fcnkey=	"MNOPQRSTUVWX"
				"YZabcdefghij"
				"klmnopqrstuv"
				"wxyz@[\\]^_`{"
				"HAI*DEC*FBGL";
	char * p;

	assert(s <= MAXSCREENS);
	assert(s >= 0);
			
	assert(screens[s] == 0);
	screens[s] = new CScreen(numRows,numCols);
	
	screens[s]->session = s;
	time(&screenTime[s]);
	ActivateScreen(s);
	
	AnsiPrintf("\33[2J\33[36;44;1m");
	AnsiPrintf(" %-77.77s\n",copyright);
	AnsiPrintf(" Terminal %2d/%2d, Id %-8.8s, Screen %02dx%02d, %-33.33s \n",
			   s+1,MAXSCREENS,TermId(),numCols,numRows,"$Revision: 1.21 $");

	AnsiPrintf("\n\33[40;32;0m");
	
	for (p = fcnkey,i = 0 ; i < 60 ;i++)
		AnsiPrintf("\33Q%d \33[%c ",i,*p++);

	Outstr("\33Q55 + \33Q51 - \33Q60 \177 ");

	for (i = 0 ; i < CScreen::FCNKEY ; i ++)
		if (scancode[i] == 0x68)
			break;

	/*
	 *	ALT-Fx per multi screen
	*/
	for (j = i ; j < i + 10 ; j ++)
	{
		AnsiPrintf("\33Q%d \33[%dz ",j,j-i+1);

	}

	//
	// Tasti funzione Linux
	if (terminalMode == Linux || terminalMode == Xterm)
	{
		AnsiPrintf("\33Q50 \33[5~ ");	// Page up
		AnsiPrintf("\33Q58 \33[6~ ");	// Page Down
		AnsiPrintf("\33Q59 \33[2~ ");	// Insert
		AnsiPrintf("\33Q60 \33[3~ ");	// delete
	}
	
	//
	// Tasti funzione vt100
	//
	if (terminalMode == Vt100)
	    	for (i = 0 ; i < 4 ; i++)
			AnsiPrintf("\33Q%d \33O%c ",i,i+'P');

	/*
	 * Tasti funzione xterm
	 */
	if (terminalMode == Xterm)
	{
		AnsiPrintf("\33Q49 \33OA "); // UP
		AnsiPrintf("\33Q57 \33OB "); // DOWN
		AnsiPrintf("\33Q54 \33OC "); // RIGHT
		AnsiPrintf("\33Q52 \33OD "); // LEFT
	}

	curScr->state = adapter->GetState();
	StatusLine("~screen %d",s+1);
	
	if (s < MAXSCREENS)
	{
		ConnectScreen(s);
		//KeybInsert(13);
	}
}

void CAnsi::ResetScreen()
{
	VideoClear(0,0,numRows,numCols);
	curScr->curPos = 0;
	ProcessAttribute(0,1);
	curScr->Default();
	Locate();
}

void CAnsi::DeactivateScreen(int s)
{
	if (s != MAXSCREENS)
		device->Deactivate(s);
}

void CAnsi::ActivateScreen(int s)
{
	int wrapped = 0;
	assert(s <= MAXSCREENS);
	assert(s >= 0);


	adapter->HideCursor();
	curScr = screens[s];
	assert(curScr != 0);

	if (!curScr->initialized)
	{
		CalcAttribute();
		SetColor();
		curScr->curPos = 0;
		curScr->Print(' ',screenSize);
		curScr->curPos = 0;
		curScr->initialized = 1;
	}

	ShowStatusLine();

	wrapped = curScr->wrapped;
	VideoPut(0,screenSize,curScr->buffer);

	adapter->SetColor(curScr->curColor);
	adapter->Locate(curScr->curPos);
	
	if (curScr->cursorOn > 0)
		adapter->ShowCursor();

	curScr->wrapped = wrapped;
}

char * CAnsi::TermId()
{
	switch (terminalMode)
	{
		case	Linux:
			return "Linux";

		case	Vt100:
			return "Vt100";

		case	Ansix364:
			return "Scoansi";

		case	Xterm:
			return "xterm";
			
		default:
			return "dump";
	}

//	return 0;
}

void CAnsi::ConnectScreen(int s)
{
	int c;


	if (!device->IsConnected(s))
	{
		startConnect[s] = ::GetTickCount();
		endConnect[s]	= 0;
	}
	
	while (device->Connect(s,TermId(),numRows,numCols))
	{
		AnsiPrintf("\33[41;37m Unable to connect strike any key to retry (ESC exit) \33[0m");
		c = KeybRead();
		AnsiPrintf("\n");
		if (c == 27)
		{
			flagExit = 1;
			break;
		}
	}
	

}

CAnsi::CAnsi(CAdapter * a,CDevice * d,EMode mode,int pselect,int dbuffer)
{
	
	assert(CScreen::FCNKEY == sizeof(scancode));
	
	int i;

	terminalMode = mode;
	selectPrinter = pselect;
	doubleBuffer = dbuffer;
	if (!defaultAnsi)
		defaultAnsi = this;
	
	deviceRead = deviceWrite = 0;
	adapter = a;
	device = d;
	flagExit = 0;
	flagSwitch = 1;
	fdCapture = -1;
	fdFromFile = -1;
	
	strcpy(printerName,"lpt1");
	
	for (i = 0 ; i < MAXSCREENS+1; i++)
		screens[i] = 0;

	adapter->ShowCursor();
	numRows = adapter->Rows();
	numCols = adapter->Cols();

	if (terminalMode == Xterm)
		numRows -= 1;
	
	
	screenSize = numCols * numRows;
	tmpBuffer = new SCHAR[screenSize];
	assert(tmpBuffer != 0);
	copyText = new char[screenSize+numRows];
	copyText[0] = 0;
}

void CAnsi::Macro()
{
	fdFromFile = open("ansi.cap",O_RDONLY|O_BINARY);
	assert(fdFromFile >= 0);
}

void CAnsi::Capture()
{
	if (fdCapture >= 0)
	{
		//AnsiPrintf("\nEnd Capture\n");
		close(fdCapture);
		fdCapture = -1;
	}
	else
	{
		//AnsiPrintf("\nStart Capture\n");
		fdCapture = open("ansi.cap",
				 O_BINARY|O_WRONLY|O_CREAT|O_APPEND,0664);
	}
}

CAnsi::~CAnsi()
{
	int i;

	if (fdFromFile >= 0)
		close(fdFromFile);
	if (fdCapture >= 0)
		close(fdCapture);
	
	for (i = MAXSCREENS ;i >= 0;  i--)
	{
		if (screens[i] != 0)
		{
			if (i != MAXSCREENS)
				device->Disconnect(i);
			delete screens[i];
		}
	}

	delete tmpBuffer;
}

void CAnsi::Main()
{
	unsigned key;
	char buffer[1024];
	int i;
	int flagWork = 1;
	clock_t lastClock=0,curClock;
	
	for (;;)
	{
		
		if (flagExit)
		{
			flagExit = 0;
			DeleteScreen();
			if (ChangeScreen() == -1)
				break;
		}

		i = 0;
		while (KeybStatus() && i < sizeof(buffer))
		{
			key = KeybRead();
			buffer[i++] = (char )key;
		}

		
		if (i)
		{
			flagWork = 1;
			Write(buffer,i);
		}

		if ((i = Read(buffer,sizeof(buffer))) < 0)
		{
			i = 0;
			flagExit = 1;
		}
		
		if (i)
		{
			flagWork = 1;
			if (fdCapture >= 0)
				write(fdCapture,buffer,i);
			Outstr(buffer,i);
		}

		if (flagWork)
			lastClock = clock();
		else
		{
			curClock = clock();
			/*if ((curClock - lastClock) > (CLOCKS_PER_SEC*60))
			{
				Sleep(100);
			}
			else*/
			if ((curClock - lastClock) > (CLOCKS_PER_SEC*1))
			{
				Sleep(100);
			}
		}
		
		flagWork = 0;

		
	}
}

CAnsi * CAnsi::defaultAnsi;
void AnsiPutchar(char c)
{
	CAnsi::defaultAnsi->Putchar(c);	
}

void AnsiPrintf(char *format,...)
{
	va_list arglist;
	char buffer[256];
	char *p;
	
	va_start( arglist, format );
	vsprintf( buffer,format, arglist );
	va_end( arglist );
	assert(strlen(buffer) < sizeof(buffer));

	for (p = buffer ; *p ; p++)
	{
		if (*p == '\n')
			CAnsi::defaultAnsi->Putchar('\r');
		CAnsi::defaultAnsi->Putchar(*p);
	}
}

void AnsiError(char *format,...)
{
	va_list arglist;
	char buffer[256];
	char *p;
	
	AnsiPrintf("\n\33[41;37m ");
	
	va_start( arglist, format );
	vsprintf( buffer,format, arglist );
	va_end( arglist );
	assert(strlen(buffer) < sizeof(buffer));

	for (p = buffer ; *p ; p++)
	{
		if (*p == '\n')
			CAnsi::defaultAnsi->Putchar('\r');
		CAnsi::defaultAnsi->Putchar(*p);
	}

	AnsiPrintf(" \33[31;47m Press any key to continue ");
	CAnsi::defaultAnsi->KeybRead();
	AnsiPrintf("\33[0m\n");
}

int CAnsi::ChangeScreen()
{
	int i;
	
	for (i = MAXSCREENS  - 1 ;i >= 0;  i--)
	{
		if (screens[i] != 0)
		{
			SwitchScreen(i);
			return i;
		}
	}

	return -1;
}

void CAnsi::HideCursor()
{
			
	if (--curScr->cursorOn == 0)
		adapter->HideCursor();
}

void CAnsi::ShowCursor()
{
	if (++curScr->cursorOn == 1  )
		adapter->ShowCursor();

}


void CAnsi::SetPrinterName(char * p)
{
	strcpy(printerName,p);
}

void CAnsi::SetSwitch(int flag)
{
	flagSwitch = flag;
}

void CAnsi::Locate()
{
	adapter->Locate(curScr->curPos < screenSize  ? curScr->curPos : screenSize -1);
}
