/**
 *
 * $Id: devtel.cpp,v 1.6 2005/05/18 11:07:55 mviara Exp $
 * $Author: mviara $
 * $Name:  $
 *
 * Device telnet.
 * 
 * $Log: devtel.cpp,v $
 * Revision 1.6  2005/05/18 11:07:55  mviara
 * Added support for  RFC1093 (Windows size)
 *
 * Revision 1.5  2004/12/06 07:31:38  mviara
 * Bug fixed in ESC 1 K
 *
 * Revision 1.4  2004/11/10 18:40:17  mviara
 * Version 2.01
 *
 * Revision 1.3  2004/07/10 14:58:39  mviara
 * Corret error in display host lost.
 *
 * Revision 1.2  2004/07/10 13:59:58  mviara
 * Add Supporto for connection time.
 *
 * Revision 1.1.1.1  2004/07/10 12:33:22  mviara
 * Migrated from RCS
 *
 *
 * 
 */

#include "ansi.hpp"

static char rcsinfo[]= "$Id: devtel.cpp,v 1.6 2005/05/18 11:07:55 mviara Exp $";

int CDeviceTelnet::CheckSocket()
{
	if (WSAGetLastError() != WSAEWOULDBLOCK)
	{
		AnsiError("Lost connetion with %s",hostName);
		return 1;
	}

	return 0;
}

int CDeviceTelnet::IsConnected(int code)
{
	return telnetSession[code].isValid;
}

int CDeviceTelnet::Connect(int port,char *id,int _row,int _col)
{
	SOCKET fd;
	SOCKADDR_IN sockAddr;
	char  host [256];
	int telnetPort = 23;
	int proxyPort = 8080;
	char proxy[256];
	char remote[256];
	int remotePort;
	
	termType = id;
	session = port;
	row = _row;
	col = _col;
	curSession  = telnetSession + session;
	
	if (curSession->isValid)
		return 0;

	strcpy(host,hostName);
     
	fd = socket(PF_INET,SOCK_STREAM,0);
	

	if (fd == INVALID_SOCKET)
	{
		AnsiError("Telnet : socket()");
		exit(1);
	}


	char *p = strchr(host,':');
	if (p != NULL)
	{
		*p = 0;
		telnetPort = atoi(p+1);
	}
	
	AnsiPrintf("Connecting to %s:%d",host,telnetPort);

	strcpy(remote,host);
	remotePort = telnetPort;
	
	if (flagProxy)
	{
		strcpy(proxy,proxyName);
		p = strchr(proxy,':');
		if (p != NULL)
		{
			*p = 0;
			proxyPort = atoi(p+1);
		}
		AnsiPrintf(" Proxy %s:%d",proxyName,proxyPort);
		strcpy(remote,proxy);
		remotePort = proxyPort;
	}
	
	memset(&sockAddr,0,sizeof(sockAddr));	
	sockAddr.sin_family = AF_INET;
	sockAddr.sin_addr.s_addr = inet_addr(remote);

	if (sockAddr.sin_addr.s_addr == INADDR_NONE)
	{
		LPHOSTENT lphost;
		lphost = gethostbyname(remote);
		if (lphost != NULL)
			sockAddr.sin_addr.s_addr = ((LPIN_ADDR)lphost->h_addr)->s_addr;
		else
		{
			AnsiError("Host not found %s",host);
			closesocket(fd);
			return 1;
		}
	}

	sockAddr.sin_port = htons((u_short)remotePort);
	
	if (connect(fd,(SOCKADDR*)&sockAddr,sizeof(sockAddr)))
	{
		AnsiError("connect() a %s:%d",remote,remotePort);
		closesocket(fd);
		return 1;
	}
	else
	{
		//AnsiPrintf("\n");
		unsigned long value = 1;
		ioctlsocket(fd,FIONBIO,&value);
		telnetSession[session].isValid = 1;
		telnetSession[session].socket = fd;
		curSession->echoTX = 0;	// bug
		curSession->echoRX = 0;	// bug

		curSession->numBufferIn =
	        curSession->readBufferIn =0;

		// Se presente il proxy manda l'indirizzo del server
		if (flagProxy)
		{
			char buffer[256];
			printf("send relay\n");
			sprintf(buffer,"relay %s %d\n",host,telnetPort);
			Write(buffer,strlen(buffer));
		}

		return 0;
	}
	
	
}

void CDeviceTelnet::Deactivate(int)
{
}

void CDeviceTelnet::Disconnect(int code)
{
	if (telnetSession[code].isValid)
	    closesocket(telnetSession[code].socket);
	
	telnetSession[code].isValid = 0;
}

CDeviceTelnet::CDeviceTelnet(char * host)
{
	WORD version;
	WSADATA data;

#ifndef MAKEWORD
#define	MAKEWORD(n1,n2)	(n1 << 8 | n2)
#endif
	
	version = MAKEWORD(1,1);	// versione winsock 1.1
	
	WSAStartup(version,&data);
	
	for (int i = 0 ; i < CAnsi::MAXSCREENS ; i++)
		telnetSession[i].isValid = 0;

	strcpy(hostName,host);
	session = 0;
	
}

char * CDeviceTelnet::Name()
{
	return "Telnet $Revision: 1.6 $";
}
	


void CDeviceTelnet::Send3(int n1,int n2,int n3)
{
	unsigned char buffer[3];
	
	buffer[0] = (unsigned char)n1;
	buffer[1] = (unsigned char)n2;
	buffer[2] = (unsigned char)n3;
	WriteRaw(buffer,3);
}

int CDeviceTelnet::WriteRaw(unsigned char * p,int len)
{
	int n;

	assert(telnetSession[session].isValid);

	n =  send(telnetSession[session].socket,(char*)p,len,0);

	if (n != len)
	{
		CheckSocket();
		return n;
	}
	
	return n;
	
}

int CDeviceTelnet::Write(char * p,int len)
{
	int i;
	
	for (i = 0 ; i < len ; i++)
	{
		if (curSession->echoTX)
			AnsiPutchar(p[i]);
	}

	WriteRaw((unsigned char*)p,i);
	
	return len;
}

void CDeviceTelnet::WONTReceived()
{
	int c;

	c = RecvChar();
	//AnsiPrintf("WONT %03d ",c);
	
	switch (c)
	{
		default:
			break;
		case	TELOPT_ECHO:
			if (curSession->echoRX)
			{
				Dont(TELOPT_ECHO);
				curSession->echoRX = 0;
			}
			break;
			
	}
}

void CDeviceTelnet::SBReceived()
{
	//extern int vt100Mode;
	
	int c,l;
	char buffer[256];
	//AnsiPrintf("SB");
	
	switch (c = RecvChar())
	{
		case	TELOPT_TTYPE:
			if (RecvChar() == 1)
			{
				sprintf(buffer,"%c%c%c%c%s%c%c",
					IAC,SB,TELOPT_TTYPE,0,
					termType,IAC,SE);
				
				l = 6 + strlen(termType);

				WriteRaw((unsigned char *)buffer,l);
			}
			break;
			
	}
}

void CDeviceTelnet::DONTReceived()
{
	int c;
	c = RecvChar();
	
	//AnsiPrintf("DONT %03d",c);	
	switch (c )
	{
		default:
		case	TELOPT_ECHO:
			if (curSession->echoTX)
			{
				Wont(TELOPT_ECHO);
				curSession->echoTX = 0;
			}
			break;
	}	

}

void CDeviceTelnet::WILLReceived()
{
	int c;

	c =RecvChar();
	//AnsiPrintf("WILL %03d",c);
	
	switch (c )
	{
		case	TELOPT_ECHO:
			if (!curSession->echoRX)
			{
				Do(TELOPT_ECHO);
				curSession->echoRX = 1;
			}
			break;
		default:
			
			Dont(c);
			break;
	}	
}

void CDeviceTelnet::DOReceived()
{
	int c;
	char buffer[256];

	c = RecvChar();
	//AnsiPrintf("DO %03d",c);
	
	switch (c)
	{
		default:
			Wont(c);
			break;
		case	TELOPT_ECHO:
			if (!curSession->echoTX)
			{
				Will(TELOPT_ECHO);
				curSession->echoTX = 1;
			}
			break;
			
		case	TELOPT_TTYPE:
				Will(c);
				break;
				
		case	TELOPT_NAWS:
				Will(c);

				sprintf(buffer,"%c%c%c%c%c%c%c%c%c",
						IAC,SB,c,0,col,0,row,IAC,SE);
				WriteRaw((unsigned char*)buffer,9);
				break;
			
	}
}

int CDeviceTelnet::IACReceived()
{
	int c;
	c = RecvChar();

	switch (c)
	{
		case	IAC:
			return 0;
		
		case	DO:
			DOReceived();
			break;
		case	WILL:
			WILLReceived();
			break;
		case	WONT:
			WONTReceived();
			break;
		case	DONT:
			DONTReceived();
			break;
		case	SB:
			SBReceived();
			break;
	}

	return 1;
}

int CDeviceTelnet::RecvChar()
{
	int c;
	
	assert(telnetSession[session].isValid);
	

	//
	// Se ci sono dei caratteri nel buffer di ingresso torna subito
	//
	if (curSession->readBufferIn < curSession->numBufferIn)
	{
		
		c =  curSession->bufferIn[curSession->readBufferIn++] & 0xff;
		return c;
	}

	
	curSession->numBufferIn = recv(curSession->socket,
				       curSession->bufferIn,
				       sizeof(curSession->bufferIn),0);
	curSession->readBufferIn = 0;

	
	if (curSession->numBufferIn  < 0)
	{
		curSession->numBufferIn = 0;
		if (CheckSocket())
			return -2;
		else
			return -1;

	}
	else	if (curSession->readBufferIn < curSession->numBufferIn)		
			return curSession->bufferIn[curSession->readBufferIn++] & 0xff;
		else
			return -1;
}
	
int CDeviceTelnet::Read(char * p,int maxlen)
{
	int c,len;

	assert(telnetSession[session].isValid);
	assert(maxlen >= 0);

	len = 0;
	
	for (;;)
	{
		c = RecvChar();

		switch (c)
		{
			case	IAC:
					if (IACReceived())
						break;
			default:
				
					*p++ = (char)(c & 0xff);
					if (++len >= maxlen)
						return len;
					break;
			case	-1:
					return len;
			case	-2:
				return -1;
		}
	}
}

CDeviceTelnet::~CDeviceTelnet()
{
	for (int i = 0 ; i < CAnsi::MAXSCREENS ; i++)
		if (telnetSession[i].isValid)
			closesocket(telnetSession[i].socket);
	
}


